#######################################################################################################
# file: analyze_pieces.py
#
# This python program reads BurrTools files and classifies pieces in the "Rhombic Tetrahedra" geometry
# Steps in order to use this program:
# 1) a) Using BurrTools (I use 0.6.3), create a new puzzle in the "Rhombic Tetrahedra" geometry.
#    b) Create a rhombic dodecahedron from 24 variable voxels (on a 2x2x2 grid),
#       and a tetra from 2 fixed voxels.  Label these two pieces and they will be ignored
#       by this program (desirable).
# Note: A sample file saved at this point is included, it is called "RawTetra.xmpuzzle",
#       you can skip steps 1a-1b by opening this file.
#    c) Now create a puzzle with the variable RD as target shape, and include as many tetras
#       as you wish.
#    d) After increasing the "Limit" (so you do not miss any shapes), solve the puzzle.
#    e) Import the generated pieces into BurrTools using File -> Import Assms.
#       Just select "Continue" in the dialog box as you probably want the defaults
#       (drop disconnected shapes, remove identical shapes).
#       Note that when these shapes are imported, some of them may be on a grid larger than 2x2x2.
#       This program looks for this and should be able to handle it.
#    f) Repeat 1d and 1e with a different number of tetras, if you want.
#    g) Save the puzzle under a name, say "tetra.xmpuzzle".
#
# 2) a) Drop into a unix shell, and rename the BurrTools file via "mv tetra.xmpuzzle tetra.gz".
#    b) Execute "gunzip tetra.gz", the (XML text file) will now be called "tetra", look at it!
#
# 3) a) Run this python program after setting the options as you desire
#       (the main options are input_file, output_file, include_mirrors,
#        export_categories and verbosity).
#
# 4) a) The BurrTools output file is called output_file, in this case "tetra_conv".
#    b) Compress this file via "gzip tetra_conv".
#    c) Rename the file via "mv tetra_conv.gz tetra_conv.xmpuzzle"
#
# 5) a) Open "tetra_conv.xmpuzzle" in BurrTools and see that the pieces have been classified.
#    b) Select "Status" to see a table of all the pieces.  If include_mirrors=1, the python program
#       will mirror every piece, even those pieces which are symmetric.
#       These pieces may be duplicated and can be removed at this point by clicking the check box.
#
# Written by: George Bell
# Last mod:   February 5th, 2014
#######################################################################################################

# Read in helper routines for reading in BurrTools XML files
from burrtools_helper import *

TETRAS = 24 # Do not change this unless you have a good reason!
input_file = "tetra" # Input BurrTools XML file
output_file = "tetra_conv" # Output BurrTools XML file

# U = Unremovable
# S = Slides out in one direction only
# F = slides out in more than one direction (Fall out)
# D = Defective (piece is not connected, or its complement is not connected)
piece_categories = ["U", "S", "F", "D"] # Do not change these

# Set to 1 to include a mirror image of each piece
include_mirrors = 1

# Export only pieces in these categories
# Usually we don't want to export pieces of type F or D
export_categories = [1,1,0,0]

# If this is set to 1, lots of info on each piece is printed out
verbosity = 0

# Determines if a piece AND its complement are connected by adding connected faces
def badPiece(faces):
  size0 = 0
  for i in range(TETRAS):
    if faces[i]==0: size0 = size0 + 1
  if size0<2: return 0
  if size0>22: return 0
  size1 = TETRAS-size0 # The size of the complemented piece

# Find a tetra in the piece and its complent to act as seeds
  for i in range(TETRAS):
    if faces[i]==0:
      s0 = set([i+1])
      break
  for i in range(TETRAS):
    if faces[i]==1:
      s1 = set([i+1])
      break
  
# Now add connected faces until there are no more to add
  add_faces = [1]
  while len(add_faces)>0:
    add_faces = []
    for j in s0:
      for k in range(len(adjacent[j-1])):
         adj_face = adjacent[j-1][k]
         if faces[adj_face-1]==0 and (adj_face not in s0): add_faces.append(adj_face)
    for k in range(len(add_faces)): s0.add(add_faces[k])

# Add connected faces to the complement until there are no more to add
  add_faces = [1]
  while len(add_faces)>0:
    add_faces = []
    for j in s1:
      for k in range(len(adjacent[j-1])):
         adj_face = adjacent[j-1][k]
         if faces[adj_face-1]==1 and (adj_face not in s1): add_faces.append(adj_face)
    for k in range(len(add_faces)): s1.add(add_faces[k])

# If not all the tetras are accounted for, the piece or the complement are not connected
  bad = 0
  if len(s0)!=size0: bad = bad + 1
  if len(s1)!=size1: bad = bad + 2
  s0 = set([])
  s1 = set([])
  return bad

# Performs roations of the RD to check all possible sliding directions.
# This will perform nz rotations about the z-axis, followed by
# ny rotations about the y-axis, followed by nx rotations about the x-axis.
def performRotations(nz, ny, nx, faces):
  new_faces = []
  for i in range(TETRAS): new_faces.append(0)

# Perform z rotation(s)
  for j in range(nz):
    for i in range(TETRAS): new_faces[i] = faces[rotz[i]-1]
    for i in range(TETRAS): faces[i] = new_faces[i]

# Perform y rotation(s)
  for j in range(ny):
    for i in range(TETRAS):
      new_faces[i] = faces[roty[i]-1]
    for i in range(TETRAS):
      faces[i] = new_faces[i]

# Perform x rotation(s)
  for j in range(nx):
    for i in range(TETRAS):
      new_faces[i] = faces[rotx[i]-1]
    for i in range(TETRAS):
      faces[i] = new_faces[i]

  new_faces = []
  return
    
# Determines whether two pieces can move apart along this direction
def canMove(faces, block):
  blocked = 0 # assume we are free to move until we find a tetra in the way
  for i in range(TETRAS):
    if faces[i]==1:
      nbl = len(block[i])
      for j in range(0,nbl):
        if faces[block[i][j]-1]==0: blocked = 1

  if blocked==1: return 0
  return 1

# An RD in BurrTools lives in a 10x10x10 grid, but only these 48 voxels can be nonzero.
# Each pair of numbers is for the same tetra.
rd_locs = [234, 235, 243, 246, 253, 256, 264, 265, 324, 325, 342, 347,
           352, 357, 374, 375, 423, 426, 432, 437, 462, 467, 473, 476,
           523, 526, 532, 537, 562, 567, 573, 576, 624, 625, 642, 647,
           652, 657, 674, 675, 734, 735, 743, 746, 753, 756, 764, 765]

# A dictionary which maps BurrTools voxel locations to face numbers of the RD.
# It assumes the BT piece is in a 2x2x2 box, which an RD normally is.
bt2face = dict(bt764=1,  bt765=1,  bt746=2,  bt756=2,  bt734=3,  bt735=3,  bt743=4,  bt753=4,
               bt674=5,  bt675=5,  bt647=6,  bt657=6,  bt624=7,  bt625=7,  bt642=8,  bt652=8,
               bt476=9,  bt576=9,  bt467=10, bt567=10, bt437=11, bt537=11, bt426=12, bt526=12,
               bt423=13, bt523=13, bt432=14, bt532=14, bt462=15, bt562=15, bt473=16, bt573=16,
               bt374=17, bt375=17, bt347=18, bt357=18, bt324=19, bt325=19, bt342=20, bt352=20,
               bt264=21, bt265=21, bt246=22, bt256=22, bt234=23, bt235=23, bt243=24, bt253=24)

# Adjacent faces for face 1,2,3,4, ... 24
# The first face is always the other half of the pair
# Used to determine if a piece is connected ...
adjacent = [[ 5, 2, 4],[ 6, 1, 3],[ 7, 2, 4],[ 8, 1, 3],[ 1, 9,16],[ 2,10,11],
            [ 3,12,13],[ 4,14,15],[10, 5,17],[ 9, 6,18],[12, 6,18],[11, 7,19],
            [14, 7,19],[13, 8,20],[16, 8,20],[15, 5,17],[21, 9,16],[22,10,11],
            [23,12,13],[24,14,15],[17,22,24],[18,21,23],[19,22,24],[20,21,23]]

# Rotations, showing the new positions of all tetras
rotx = [ 7,12,19,13,   3,11,23,14,
         2, 6,18,22,  24,20, 8, 4,
         1,10,21,15,   5, 9,17,16]

roty = [10,18,11, 6,   9,22,12, 2,
        17,21,23,19,   7, 3, 1, 5,
        16,24,13, 4,  15,20,14, 8]

rotz = [ 4, 1, 2, 3,   8, 5, 6, 7,
        15,16, 9,10,  11,12,13,14,
        20,17,18,19,  24,21,22,23]

# Blocking pieces, to see if two pieces can come apart
# Along z-axis (4-fold symmetry)
block1 = [ [],  [],  [],  [],
          [1], [2], [3], [4],
          [1,5], [2,6], [2,6], [3,7],
          [3,7], [4,8], [4,8], [1,5],
          [1,5,9,16], [2,6,10,11], [3,7,12,13], [4,8,14,15],
          [1,5,9,16,17], [2,6,10,11,18], [3,7,12,13,19], [4,8,14,15,20]]

# Along a diagonal axis (through a point of 3-fold symmetry)
block2 = [ [], [], [2], [1],
           [], [], [2,3], [1,4],
           [], [], [6], [6,11],
           [2,3,7], [1,4,8], [5,16], [5],
           [9], [10], [6,11,12], [5,16,15],
           [9, 17], [10,18], [10,18,22], [9,17,21]]

piece_counts = []
for i in range(len(piece_categories)): piece_counts.append([0,0,0,0,0,0,0,0,0,0,0,0,0,0])

piece_faces = []
rot_faces = []  # To store rotations
new_faces = []  # To store rotations
for i in range(TETRAS):
  piece_faces.append(0) # The 24 faces, initialized to zero
  rot_faces.append(0)
  new_faces.append(0)

# The number of directions each piece can come out
# 0 indicates an unremovable piece
# 1 indivates a sliding piece
# >1 indicates a falling out piece
piece_dirs = 0

done = 0
piece_no = 0
piece_size = 0

fi = open(input_file)
fo = open(output_file,"w")

while not done:
  line = fi.readline()
  if not line: break

# Check for the wrong srid type
  test = line.find('gridType')
  if test>0:
    grid = getValueInt('type',line)
    if grid!=3:
      print "This puzzle has the wrong grid type, it must be 'Rhombic Tetrahedra'"
      break

# Look for a piece
  test = line.find('voxel ')
  if test>0:
    piece_no = piece_no + 1
    mod_ok = 1

    myname = getValueStr('name',line)

    myx = getValueInt('x',line)
    myy = getValueInt('y',line)
    myz = getValueInt('z',line)

    data = getString('voxel',line)
#    print data,myx,myy,myz

# Automatically skip pieces that are labeled
    if myname!='':
      mod_ok = 0
      if verbosity==1: print "Piece ",piece_no," ("+myname+") was ignored"

# If it is not 2x2x2, we must cut it down to size
    newdata = ""
    i = 0
    if len(data)!=1000: 
      for curz in range(0,myz):
        for cury in range(0,myy):
          for curx in range(0,myx):
            if curx<10 and cury<10 and curz<10: newdata = newdata + data[i:i+1]
            i = i + 1
      if mod_ok>0: data = newdata
#      print "Length is now: ",len(data)

      if mod_ok==0:
        fo.write('  <voxel x="10" y="10" z="10" type="0">')
        fo.write(data)
        fo.write('</voxel>\n')
    else:
      if mod_ok==0:
        fo.write(line)

    if mod_ok==1: # Continue and modify the piece
# get the mirror image (in newdata)
      newdata = data
      for i in range(0,len(rd_locs),2):
        loc0 = rd_locs[i];
        loc1 = rd_locs[i+1];

        newchar = data[loc1:loc1+1]
        newdata = newdata[:loc0] + newchar + newdata[loc0+1:]

        newchar = data[loc0:loc0+1]
        newdata = newdata[:loc1] + newchar + newdata[loc1+1:]

      for i in range(TETRAS): piece_faces[i] = 0
      piece_size = 0
      piece_category = 0
      piece_dirs = 0

      for i in range(0,len(rd_locs)):
        loc = rd_locs[i];
        locstr = 'bt'+str(loc)
        curchar = data[loc:loc+1]

        if locstr in bt2face:
          faceno = bt2face[locstr]
        else:
          faceno = 0
          print "Error, BurrTools location ",loc," doesn't map to any face??"

        newchar = curchar
        if curchar=="#": 
          if piece_faces[faceno-1]==0: piece_size = piece_size + 1
          piece_faces[faceno-1] = 1

      if verbosity==1:
        print "Piece ",piece_no," has size ",piece_size," and faces ",
        for i in range(TETRAS):
          if piece_faces[i]==1: print i+1,
        print
      if badPiece(piece_faces)>0: piece_category = 3

# Now we determine in how many directions the piece can slide apart
      for i in range(TETRAS): rot_faces[i] = piece_faces[i]

      piece_dirs = piece_dirs + canMove(rot_faces, block2)
      performRotations(1, 0, 0, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block2)
      performRotations(1, 0, 0, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block2)
      performRotations(1, 0, 0, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block2)
      performRotations(1, 2, 0, rot_faces)

      piece_dirs = piece_dirs + canMove(rot_faces, block2)
      performRotations(1, 0, 0, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block2)
      performRotations(1, 0, 0, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block2)
      performRotations(1, 0, 0, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block2)
      performRotations(1, 2, 0, rot_faces)

# Should be back on the identity at this point

      piece_dirs = piece_dirs + canMove(rot_faces, block1)
      performRotations(0, 1, 0, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block1)
      performRotations(0, 1, 0, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block1)
      performRotations(0, 1, 0, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block1)
      performRotations(0, 1, 1, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block1)
      performRotations(0, 0, 2, rot_faces)
      piece_dirs = piece_dirs + canMove(rot_faces, block1)

#      print "Number of directions of movement: ",piece_dirs

      if piece_category==0 and piece_dirs>1: piece_category = 2
      if piece_category==0 and piece_dirs==1: piece_category = 1

# Determine the name of this piece
      piece_counts[piece_category][piece_size] = piece_counts[piece_category][piece_size] + 1
      piece_name = str(piece_size)+piece_categories[piece_category]+str(piece_counts[piece_category][piece_size])

# Write the new piece if it is OK

      if len(newdata)!=1000:
        print "Error in piece ",data

      elif export_categories[piece_category]==0:
        if verbosity==1: print "Not exporting this piece: ",piece_name

      else:
        fo.write('  <voxel x="10" y="10" z="10" name="'+piece_name+'" type="0">')
        fo.write(data)
        fo.write('</voxel>\n')
        if verbosity==1: print "Piece ",piece_no," was given the label ",piece_name

        if include_mirrors>0:
          fo.write('  <voxel x="10" y="10" z="10" name="'+piece_name+'m'+'" type="0">')
          fo.write(newdata)
          fo.write('</voxel>\n')

  else: # voxel not found ...
    fo.write(line) # Keep the rest of the BurrTools file unchanged

fi.close()
fo.close()

print "Total   Locking pieces:", sum(piece_counts[0]), piece_counts[0]
print "Total   Sliding pieces:", sum(piece_counts[1]), piece_counts[1]
print "Total   Falling pieces:", sum(piece_counts[2]), piece_counts[2]
print "Total Defective pieces:", sum(piece_counts[3]), piece_counts[3]

